:- module doc_optimize.

:- interface.

:- import_module smlng_data.
:- import_module io.

:- pred doc_optimize(document::in, string::in,
		io__state::di, io__state::uo) is det.

:- implementation.

:- import_module bool, list, require, std_util, exception, string, int.
:- import_module attrs, output, validator.

doc_optimize(Document, FileNameBase) -->
	doc_optimize_2(Document, FileNameBase, 1).

:- pred doc_optimize_2(document::in, string::in, int::in,
		io__state::di, io__state::uo) is det.

doc_optimize_2(Document, FileNameBase, Stage0) -->

	open_stage_file(FileNameBase, Stage0),

	{ OptDoc1 = unused_tags(Document) },

	{ validate(Document, OptDoc1, Result) },
	{ Result = no,
		true
		%ValidDoc = OptDoc1
	; Result = yes(Error),
		error(Error)
		% ValidDoc = Document
	},
	output(OptDoc1),
	close_stage_file(FileNameBase, Stage0),

	% --------------------------------------------------

	{ Stage1 = Stage0 + 1 },
	open_stage_file(FileNameBase, Stage1),

	{ OptDoc2 = compress_whitespace(OptDoc1, no) },
	output(OptDoc2),

	close_stage_file(FileNameBase, Stage1),

	% --------------------------------------------------


	{ Stage2 = Stage1 + 1 },
	open_stage_file(FileNameBase, Stage2),

	{ OptDoc3 = redundancy_elimination(OptDoc2, root_attrs) },
	output(OptDoc3),

	close_stage_file(FileNameBase, Stage2),

	% --------------------------------------------------

	{ Stage3 = Stage2 + 1 },
	open_stage_file(FileNameBase, Stage3),

	{ OptDoc4 = unused_tags(OptDoc3) },
	output(OptDoc4),

	close_stage_file(FileNameBase, Stage3),

	% --------------------------------------------------

	{ Stage4 = Stage3 + 1 },
	open_stage_file(FileNameBase, Stage4),

	{ OptDoc5 = remove_empty_docs(OptDoc4) },
	output(OptDoc5),

	close_stage_file(FileNameBase, Stage4),

	% --------------------------------------------------


	( { OptDoc5 = OptDoc1 } ->
		[]
	;
		doc_optimize_2(OptDoc5, FileNameBase, Stage4 + 1)
	).

	
	

:- pred open_stage_file(string::in, int::in, io__state::di,
	io__state::uo) is det.
open_stage_file(FileNameBase, Stage) -->
	io__tell(format("%s-%d.smlng", [s(FileNameBase), i(Stage)]), FileRes), 
	( if { FileRes = error(_) } then { throw(FileRes) } ).

:- pred close_stage_file(string::in, int::in, io__state::di,
	io__state::uo) is det.
close_stage_file(FileNameBase, Stage) -->
	io__told,
	io__tell(format("%s-%d.done", [s(FileNameBase), i(Stage)]),
		DoneRes), 
	( if { DoneRes = error(_) } then { throw(DoneRes) } ),
	io__told.

%-----------------------------------------------------------------------------%

:- func compress_whitespace(document, bool) = document.

compress_whitespace(doc(DocA, DocB), InTT)
	= doc(compress_whitespace(DocA, InTT), compress_whitespace(DocB, InTT)).
compress_whitespace(style(Style, Doc), InTT0)
	= style(Style, compress_whitespace(Doc, InTT)) :-
	( Style = tt ->
		InTT = yes
	;
		InTT = InTT0
	).
compress_whitespace(size(Size, Doc), InTT)
	= size(Size, compress_whitespace(Doc, InTT)).
compress_whitespace(colour(Colour, Doc), InTT)
	= colour(Colour, compress_whitespace(Doc, InTT)).
compress_whitespace(text(TextChars0), InTT)
	= text(TextChars) :-
	( InTT = yes,
		TextChars = TextChars0
	; InTT = no,
		SpacesOnly = list__map(ws_to_spc, TextChars0),
		TextChars = list__remove_adjacent_dups(SpacesOnly)
	).

:- func ws_to_spc(textchar) = textchar.

ws_to_spc(s(Str)) = s(Str).
ws_to_spc(cr) = spc.
ws_to_spc(lf) = spc.
ws_to_spc(tab) = spc.
ws_to_spc(spc) = spc.

%-----------------------------------------------------------------------------%

	% If the attributes are the same after a tag has been applied,
	% then the tag is uneccessary.
:- func redundancy_elimination(document, attrs) = document.

redundancy_elimination(doc(DocA, DocB), Attrs)
	= doc(redundancy_elimination(DocA, Attrs),
			redundancy_elimination(DocB, Attrs)).
redundancy_elimination(style(Style, Doc), Attrs)
	=  OptDoc :-
	NewAttrs = apply_tag(style(Style, Doc), Attrs),
	OptSubDoc = redundancy_elimination(Doc, NewAttrs),
	( NewAttrs = Attrs ->
		OptDoc = OptSubDoc
	;
		OptDoc = style(Style, OptSubDoc)
	).
redundancy_elimination(size(Size, Doc), Attrs)
	=  OptDoc :-
	NewAttrs = apply_tag(size(Size, Doc), Attrs),
	OptSubDoc = redundancy_elimination(Doc, NewAttrs),
	( NewAttrs = Attrs ->
		OptDoc = OptSubDoc
	;
		OptDoc = size(Size, OptSubDoc)
	).
redundancy_elimination(colour(Colour, Doc), Attrs)
	=  OptDoc :-
	NewAttrs = apply_tag(colour(Colour, Doc), Attrs),
	OptSubDoc = redundancy_elimination(Doc, NewAttrs),
	( NewAttrs = Attrs ->
		OptDoc = OptSubDoc
	;
		OptDoc = colour(Colour, OptSubDoc)
	).
redundancy_elimination(text(TextChars), _) = text(TextChars).

%-----------------------------------------------------------------------------%

	% If there are no items which a size tag applies to between the
	% current size tag and the next size tag then remove the current
	% size tag.
:- func unused_tags(document) = document.

unused_tags(doc(DocA, DocB))
	= doc(unused_tags(DocA), unused_tags(DocB)).
unused_tags(style(Style, Doc))
	=  style(Style, unused_tags(Doc)).
unused_tags(size(Size, Doc))
	=  OptDoc :-
	OptDoc0 = unused_tags(Doc),
	( no_docs_or_text_for_size(Doc) ->
		OptDoc = OptDoc0
	;
		OptDoc = size(Size, OptDoc0)
	).
unused_tags(colour(Colour, Doc))
	=  OptDoc :-
	OptDoc0 = unused_tags(Doc),
	( no_docs_or_text_for_colour(Doc) ->
		OptDoc = OptDoc0
	;
		OptDoc = colour(Colour, OptDoc0)
	).
unused_tags(text(TextChars)) = text(TextChars).

:- pred no_docs_or_text_for_size(document::in) is semidet.

no_docs_or_text_for_size(style(_, Doc)) :-
	no_docs_or_text_for_size(Doc).
no_docs_or_text_for_size(size(_, _Doc)).
no_docs_or_text_for_size(colour(_, Doc)) :-
	no_docs_or_text_for_size(Doc).

:- pred no_docs_or_text_for_colour(document::in) is semidet.

no_docs_or_text_for_colour(style(_, Doc)) :-
	no_docs_or_text_for_colour(Doc).
no_docs_or_text_for_colour(size(_, Doc)) :-
	no_docs_or_text_for_colour(Doc).
no_docs_or_text_for_colour(colour(_, _)).

%-----------------------------------------------------------------------------%

:- func remove_empty_docs(document) = document.

remove_empty_docs(doc(DocA, DocB))
	= doc(remove_empty_docs(DocA), remove_empty_docs(DocB)).
remove_empty_docs(style(Style, Doc))
	= OptDoc :-
	OptDoc0 = remove_empty_docs(Doc),
	( OptDoc0 = text([]) ->
		OptDoc = Doc
	;
		OptDoc = style(Style, OptDoc0)
	).
remove_empty_docs(colour(Colour, Doc))
	= OptDoc :-
	OptDoc0 = remove_empty_docs(Doc),
	( OptDoc0 = text([]) ->
		OptDoc = Doc
	;
		OptDoc = colour(Colour, OptDoc0)
	).
remove_empty_docs(size(Size, Doc))
	= OptDoc :-
	OptDoc0 = remove_empty_docs(Doc),
	( OptDoc0 = text([]) ->
		OptDoc = Doc
	;
		OptDoc = size(Size, OptDoc0)
	).
remove_empty_docs(text(TextChars)) = text(TextChars).

%-----------------------------------------------------------------------------%
