%------------------------------------------------------------------------------%
% optimize.m
% Ralph Becket <rbeck@microsoft.com>
% Fri Jul 27 01:58:09 BST 2001
% vim: ts=4 sw=4 et tw=0 wm=0 ff=unix ft=mercury
%
% The input module delivers (as part of its result) the sequence of
% attribute records for the source document.
%
% This module tries to optimize as it writes the output (NOTE: something
% may go wrong and we could end up making matters worse...  This is
% unavoidable given the timing constraints.)
%
% The optimizer keeps track of the current stack of open tags as it
% traverses the attribute record array.
%
% 1 - CLOSING. When the traversal indicates one or more of the
% following state changes:
%   B, S, I, TT from on -> off
%   EM          from on -> off when S off
%   U           from =< 3 -> lower
%   SIZE        changes
%   COLOUR      changes
% then the output has to emit the appropriate close tag (possibly
% closing others according to the open tag stack until we find
% the matching tag, the intervening tags may then have to be
% reopened immediately afterwards if they are not also turned off
% in the transition.)
%
% One alternative that may be worth considering is to instead 
% insert a PL block and fix up to match (HEURISTIC REQUIRED).
%
% 2 - U LEVELS. When the traversal indicates an increase in U from
% >= 3 then no action is necessary on this score.
%
% 3 - ROOT SIZE. When the traversal indicates a change in size from
% X to root_size, the only option is to emit a </X>.  No attempt
% should be made to open <root_size>.
%
% If the traversal indicates a change in size from root_size
% to X then no attempt should be made to close a (non-existent)
% <root_size> tag.  The open <X> tag is required, of course.
%
% 4 - ROOT COLOUR. When the traversal indicates a change in colour from
% X to root_colour, the only option is to emit a </X>.  No attempt
% should be made to open <root_colour>.
%
% If the traversal indicates a change in colour from root_size
% to X then no attempt should be made to close a (non-existent)
% <root_colour> tag.  The open <X> tag is required, of course.
%
% 5 - OPENING.  When the traversal indicates one or more of the
% following state changes:
%   B, S, I, TT from off -> on
%   EM          from off -> on when S off
%   U           from < 3 -> higher
%   SIZE        changes
%   COLOUR      changes
% then the output has to emit the appropriate open tags.  It *may* be
% wise to close some other open tags before adding the open tag.  THIS
% IS WHERE SOME KIND OF HEURISTIC IS REALLY NECESSARY.
%
% So what sort of heuristics and search strategies should we adopt?
%
% It seems to me that what we ... (beware of the arrrggghhhh.)
%
%------------------------------------------------------------------------------%

:- module optimize.

:- interface.

:- import_module input.



%------------------------------------------------------------------------------%

:- implementation.

%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%

    % [FIRST CUT]
    %
    % The optimizer works by computing the shortest tag sequence for
    % each consecutive N-sequence of attribute records and appending
    % them.  In the case of a draw the choice is made arbitrarily.
    % N is a search procedure parameter.
    %
    % The sequence generator takes the following arguments:
    % - the current attribute record;
    % - the index of the document attribute record we need to change to;
    % - the end-of-sequence index;
    % - the stack of currently open tags and the attribute records they
    % were applied to;
    % - the sequence of tokens to this point.
    %  
    % The non-determimistic generation relation works as follows:
    % - if a match has been achieved then
    %   - if we're reached the end of the sequence then return the current
    %   token sequence;
    %   - else recurse for the next index in the sequence
    % - else we still have some matching work to do:
    %   - if there are any *necessary closures*, namely
    %     * a size change to the root_size or
    %     * a colour change to the root_colour
    %     then
    %     - pop the stack (extending the tag list) until the requisite
    %     closures have been made and recurse
    %   - else
    %     - close a tag that needs closing and recurse - or
    %     - add a tag to set an attribute that needs setting - or
    %     - add a PL to close one or more tags that need closing and
    %     recurse.
    %     * If EM is in force and needs deleting then either close it
    %     or open a new EM.
    %
    % * Micro-optimization: after the document has been optmized by
    % the above procedure, we can change the innermost EM blocks that
    % do set the EM attribute to just S which saves a couple of bytes
    % in the output.
    %
    % * Once the generator has pushed a tag, there's no point in
    % it popping afterwards (it would be pointless and lead to
    % unbounded recursion).
    %
    % * Iterative deepening is the right strategy here, otherwise we're
    % going to have an even more enormous search space.



:- type ta_stack == list({tag, attrs}).



    % Find the smallest solution for the next few transformations.
    %
:- pred gen_n_cc(int, int, attrs, attrss, ta_stack, ta_stack,
            list(tokens), list(tokens)).
:- mode gen_n_cc(in, in, in, in, in, out, in, out) is cc_multi.

gen_n_cc(I, End, As, Attrss, Stk0, Stk, Tss0, Tss) :-
    nat(N),
    gen_n(I, End, As, Attrss, Stk0, Stk, N, _, Tss0, Tss).



:- pred nat(int::out) is multi.

nat(0).
nat(N + 1) :- nat(N).



    % Find a solution transforming As into Attrss elems I, I + 1, ..., End.
    %
:- pred gen_n(int, int, attrs, attrss, ta_stack, ta_stack, int, int,
            list(tokens), list(tokens)).
:- mode gen_n(in, in, in, in, in, out, in, out, in, out) is nondet.

gen_n(I, End, As, Attrss, Stk0, Stk, N0, N, Tss0, Tss) :-
    ( if I > End then
        Stk = Stk0,
        N   = N0,
        Ts  = Ts0
      else
        AsI = Attrss ^ elem(I),
        gen(As, AsI, Stk0, Stk1, N0, N1, [], Ts),
        gen_n(I + 1, End, AsI, Attrss, Stk1, Stk, N1, N, [Ts | Tss0], Tss)
    ).



    % gen(As0, As, Stk0, Stk, N0, N, Ts0, Ts)
    %
    % Work out how to change As0 to As starting with open tag-attrs stack
    % Stk0 adding no more than N0 tags to Ts0.
    %
:- pred gen(attrs, attrs, ta_stack, ta_stack, int, int, tokens, tokens).
:- mode gen(in, in, in, out, in, out, in, out) is nondet.

gen(As, AsI, Stk0, Stk, N0, N, Ts0, Ts) :-

    ( if equivalent(As, AsI) then

        Stk = Stk0,
        N   = N0,
        Ts  = Ts0

      else if closing_necessary(As, AsI) then

        det_close_tag(Stk0, Stk1, N0, N1, Ts0, Ts1, As1),
        gen(As1, AsI, Stk1, Stk, N1, N, Ts1, Ts)

      else

        (
            close_tag(Stk0, Stk1, N0, N1, Ts0, Ts1, As1),
            gen(As1, AsI, Stk1, Stk, N1, N, Ts1, Ts)
        ;
            gen_open(As, AsI, Attrss, Stk0, Stk, N0, N, Ts0, Ts)
        ;
            open_tag(pl, As, As1, Stk0, Stk1, N0, N1, Ts0, Ts1),
            gen_open(As1, AsI, Attrss, Stk1, Stk, N1, N, Ts1, Ts)
        )
    ).



:- pred det_close_tag(ta_stack, ta_stack, int, int, tokens, tokens, attrs).
:- mode det_close_tag(in, out, in, out, in, out, out) is det.

det_close_tag(Stk0, Stk, N0, N, Ts0, Ts, As) :-
    ( if   close_tag(Stk0, Stk1, N0, N1, Ts0, Ts1, As1)
      then Stk = Stk1, N = N1, Ts = Ts1, As = As1
      else throw("optimize__det_pop_tag: squawk!")
    ).



    % Try to close the top tag on the stack.
    %
:- pred close_tag(ta_stack, ta_stack, int, int, tokens, tokens, attrs).
:- mode close_tag(in, out, in, out, in, out, out) is semidet.

close_tag([{Tag, As} | Stk], Stk, N, N - 1, Ts, [close(Tag) | Ts], As) :-
    N > 0.



    % Open a new tag.
    %
:- pred open_tag(tag,attrs,attrs,ta_stack,ta_stack,int,int,tokens,tokens).
:- mode open_tag(in, in, out, in, out, in, out, in, out) is semidet.

open_tag(Tag, As0, As, Stk, [{Tag,As0} | Stk], N, N-1, Ts, [open(Tag) | Ts]) :-
    N > 0,
    As = apply_tag(Tag, As0).



    % Find the means to change one attribute record into another by
    % only opening new tags.
    %
:- pred gen_open(attrs, attrs, ta_stack, ta_stack, int, int, tokens, tokens).
:- mode gen_open(in, in, in, out, in, out, in, out) is nondet.

gen_open(As, AsI, Stk0, Stk, N0, N, Ts0, Ts) :-

    ( if equivalent(As, AsI) then

        Stk = Stk0,
        N   = N0,
        Ts  = Ts0

      else

        (
            As ^ b  = no, AsI ^ b  = yes,   Tag = b
        ;                          
            As ^ i  = no, AsI ^ i  = yes,   Tag = i
        ;                          
            As ^ tt = no, AsI ^ tt = yes,   Tag = tt
        ;                          
                 As ^ em \= AsI ^ em,       Tag = em
        ;                          
            As ^ u < AsI ^ u, As ^ u < 3,   Tag = u(0)
        ;
               As ^ size \= AsI ^ size,     Tag = size(AsI ^ size)
        ;
             As ^ colour \= AsI ^ colour,   Tag = colour(AsI ^ colour)
        ;
            (   As ^ b  = yes, AsI ^ b  = no
            ;   As ^ i  = yes, AsI ^ i  = no
            ;   As ^ tt = yes, AsI ^ tt = no
            ;   As ^ em = yes, AsI ^ tt = no
            ;   As ^ u > AsI ^ u, AsI ^ u < 3
            ),
            Tag = pl
        )
        open_tag(Tag, As, As1, Stk0, Stk1, N0, N1, Ts0, Ts1),
        gen_open(As1, AsI, Stk1, Stk, N1, N, Ts1, Ts)

    ).



% For each N-stretch, generate the shortest stretch of tokens.
% Append them all together until we're done.
% Output them interspersed with the intervening plaintext.
%
