:- module parse.

:- interface.

:- import_module smlng_data.
:- import_module io.

:- pred parse(document::out, io__state::di, io__state::uo) is det.

:- implementation.

:- import_module attrs.
:- import_module char, io, list, require, string.

parse(Document) -->
	io__read_file(Result),
	( { Result = eof },
		{ Document = text([]) }
	; { Result = error(_) },
		{ error("I/O error") }
	; { Result = ok(CharList) },
		( { parse_document(Doc, CharList, []) } ->
			{ Document = Doc }
		;
			{ error("Parse error") }
		)
	).

:- pred parse_document(document::out, list(char)::in, list(char)::out) is det.

parse_document(text([]), [], []).
parse_document(Document, [H|T], Out) :-
	parse_one_doc(DocA, [H|T], Out0),
	( DocA = text([]) ->
		Document = DocA,
		Out = Out0
	;
		parse_document(DocB, Out0, Out),
		( DocB = text([]) ->
			Document = DocA
		;
			Document = doc(DocA, DocB)
		)
	).

:- pred parse_one_doc(document::out, list(char)::in, list(char)::out) is det.

parse_one_doc(Document) -->
	( parse_tag(Tag) ->
		{ Document = Tag }
	; 
		parse_text(Text),
		{ Document = Text }
	).

:- pred parse_tag(document::out, list(char)::in, list(char)::out) is semidet.

parse_tag(Tag) -->
	[('<')],
	( 
		['B', ('>')],
		parse_document(Doc),
		[('<'), '/', 'B', ('>')],
		{ Tag = style(b, Doc) }
	;
		['E', 'M', ('>')],
		parse_document(Doc),
		[('<'), '/', 'E', 'M', ('>')],
		{ Tag = style(em, Doc) }
	;
		['I', ('>')],
		parse_document(Doc),
		[('<'), '/', 'I', ('>')],
		{ Tag = style(i, Doc) }
	;
		['P', 'L', ('>')],
		parse_document(Doc),
		[('<'), '/', 'P', 'L', ('>')],
		{ Tag = style(pl, Doc) }
	;
		['S', ('>')],
		parse_document(Doc),
		[('<'), '/', 'S', ('>')],
		{ Tag = style(s, Doc) }
	;
		['T', 'T', ('>')],
		parse_document(Doc),
		[('<'), '/', 'T', 'T', ('>')],
		{ Tag = style(tt, Doc) }
	;
		['U', ('>')],
		parse_document(Doc),
		[('<'), '/', 'U', ('>')],
		{ Tag = style(u, Doc) }
	;
		['0', ('>')],
		parse_document(Doc),
		[('<'), '/', '0', ('>')],
		{ Tag = size(0, Doc) }
	;
		['1', ('>')],
		parse_document(Doc),
		[('<'), '/', '1', ('>')],
		{ Tag = size(1, Doc) }
	;
		['2', ('>')],
		parse_document(Doc),
		[('<'), '/', '2', ('>')],
		{ Tag = size(2, Doc) }
	;
		['3', ('>')],
		parse_document(Doc),
		[('<'), '/', '3', ('>')],
		{ Tag = size(3, Doc) }
	;
		['4', ('>')],
		parse_document(Doc),
		[('<'), '/', '4', ('>')],
		{ Tag = size(4, Doc) }
	;
		['5', ('>')],
		parse_document(Doc),
		[('<'), '/', '5', ('>')],
		{ Tag = size(5, Doc) }
	;
		['6', ('>')],
		parse_document(Doc),
		[('<'), '/', '6', ('>')],
		{ Tag = size(6, Doc) }
	;
		['7', ('>')],
		parse_document(Doc),
		[('<'), '/', '7', ('>')],
		{ Tag = size(7, Doc) }
	;
		['8', ('>')],
		parse_document(Doc),
		[('<'), '/', '8', ('>')],
		{ Tag = size(8, Doc) }
	;
		['9', ('>')],
		parse_document(Doc),
		[('<'), '/', '9', ('>')],
		{ Tag = size(9, Doc) }
	;
		['r', ('>')],
		parse_document(Doc),
		[('<'), '/', 'r', ('>')],
		{ Tag = colour(r, Doc) }
	;
		['g', ('>')],
		parse_document(Doc),
		[('<'), '/', 'g', ('>')],
		{ Tag = colour(g, Doc) }
	;
		['b', ('>')],
		parse_document(Doc),
		[('<'), '/', 'b', ('>')],
		{ Tag = colour(b, Doc) }
	;
		['c', ('>')],
		parse_document(Doc),
		[('<'), '/', 'c', ('>')],
		{ Tag = colour(c, Doc) }
	;
		['m', ('>')],
		parse_document(Doc),
		[('<'), '/', 'm', ('>')],
		{ Tag = colour(m, Doc) }
	;
		['y', ('>')],
		parse_document(Doc),
		[('<'), '/', 'y', ('>')],
		{ Tag = colour(y, Doc) }
	;
		['k', ('>')],
		parse_document(Doc),
		[('<'), '/', 'k', ('>')],
		{ Tag = colour(k, Doc) }
	;
		['w', ('>')],
		parse_document(Doc),
		[('<'), '/', 'w', ('>')],
		{ Tag = colour(w, Doc) }
	).

:- pred parse_text(document::out, list(char)::in, list(char)::out) is det.

parse_text(Text) -->
	parse_text_2(TextChars),
	{ Text = text(TextChars) }.

:- pred parse_text_2(list(textchar)::out,
		list(char)::in, list(char)::out) is det.

parse_text_2(TextChars) -->
	( ['\n'] ->
		parse_text_2(TextCharsRest),
		{ TextChars = [cr | TextCharsRest] }
	; ['\r'] ->
		parse_text_2(TextCharsRest),
		{ TextChars = [lf | TextCharsRest] }
	; ['\t'] ->
		parse_text_2(TextCharsRest),
		{ TextChars = [tab | TextCharsRest] }
	; [' '] ->
		parse_text_2(TextCharsRest),
		{ TextChars = [spc | TextCharsRest] }
	; [Char], { \+ (Char = ('<') ; Char = ('>')) } ->
		parse_chars(Chars),
		parse_text_2(TextCharsRest),
		{ Str = s(string__from_char_list([Char | Chars])) },
		{ TextChars = [Str | TextCharsRest] }
	;
		{ TextChars = [] }
	).

:- pred parse_chars(list(char)::out, list(char)::in, list(char)::out) is det.

parse_chars(Chars) -->
	( [Char],  { \+ special_char(Char) } ->
		parse_chars(CharsRest),
		{ Chars = [Char | CharsRest] }
	;
		{ Chars = [] }
	).

:- pred special_char(char::in) is semidet.

special_char('\n').
special_char('\r').
special_char('\t').
special_char(' ').
special_char(('<')).
special_char(('>')).
